#!/usr/bin/perl -w  # this path to Perl is your_unix_system-dependent: either /usr/bin/perl or /usr/local/bin/perl
# Author:  Dr. Ruming Li -- December 20, 2010
# Purpose: Remove duplicate sequences, pseudo-genes, identities, overlaps, similar
#          genes, and homologs, and check oligo existence and generate unique genes
# Usage:   perl orf-purger.pl input_file [output_file][all_other_options][penalties]
#------------------------------------------------------------------------------------
use File::Basename;
use strict;
use warnings;
$0=~s/^\.\///;
my $usage = "\nUsage:\n\n$0 [switches]\n\nSwitches:\n\n-i <input-file-mode>\n-f <input-gene-files>
-o <output-genes-file>\n-O <output-overlaps-file>\n-s <orf-stats-file>\n-t <orf-tags-file>
-p <orf-ids-prefix>\n-b <orf-start-codons>\n-c <orf-components>\n-m <minimum-orf-length>
-l <oligo-probe-length>\n-r <remove-similarity>\n-e <eliminate-homology>\n-h <program-usage-info>\n
Note: the spacing following a switch (-x) is optional; eg, either '-i1' or '-i 1' is acceptable.


Synopsis:


-i <input-file-mode>

   refers to a way to input sequence file(s) and has 3 options:

   1 -- input a single/multiple file(s) by entering filename(s),
   2 -- input many files by a file in which filenames are listed with one name a line, and
   3 -- get the group of files stored in a directory and input them by entering the path
        to that directory.

   Note: All files are input in the order of their names listed, which is particularly useful in
   specifying a desired input order, as required by selective input of sequence file(s). For instance,
   you wan to input files from well-curated GenBank databases first, then files from other databases.
   Especially, if one would like to take the notations/nomenclature of .gbk file entries as standards
   (used as norms), their .ffn files take precedence to be processed over other input file entries; or
   if there is a scenario where published/well-established data files should precede the draft forms.
   By designating an input order, you mean you are using all the first-input file data as a criterion
   (benchmark), which offers you one more chance to customize a sequence purgation the way you wish.

-f <input_gene_files>

   refers to the file(s) containing target gene sequences to be input.

   The input-gene-files can be input as:

   a) a list of sequence files (Mode 1) with their names separated by the comma (,) without
      spaces between (,) and name,

   b) a name of the file having the list of input file names inside (Mode 2) and must contain
      a path to that list file if it is not in the working directory, or

   c) just a directory pathname (Mode 3).

   Note: For those sequence files whose fasta header lines contain no organism-name identity, their
   filenames must end up with the cataloging/numbering notation of an organism name, such as a name
   'Clostridium_thermocellum_DSM_2360.ffn' or 'Clostridium_thermocellum_JW20.ffn'. The last symbol
   (ie, 2360 or JW20) will be used to construct an orf/oligo ID for input gene identification. This
   will also be done by specifying a desired prefix from the option <orf-ids-prefix> for your own
   orf/gene/oligo IDs. For those sequence files whose fasta header lines contain no gene symbols or
   have them but are not unique, their orf/gene IDs grabbed will be used as gene symbols, as those
   orf/gene IDs are distinct across entries. And then the orf/gene IDs will be created by the rule
   above, i.e., using the last portion (symbol) of an input filename to prefix a sequential number.
   This way also enables you to customize individual orf/gene/oligo IDs separately and selectively.

** All missing data found in input sequence files for the items gene symbol, gene locus, and gene
   product name will be represented as gene_symbol, gene_locus, and product_name, respectively, for
   the purposes of better line-up data view, best data integrity, and fasta header line formatting.

   With Mode 1, an input file name may not contain blank spaces and brackets in the unix systems.
   But Modes 2 and 3 allow files to be input whose names can contain blank spaces and brackets.

   Examples for Mode 1:

   If the sequence file(s) is (are) in the working directory from which the program is invoked,
   the input-gene-files can be input as

	> orf-purger.pl -f seq.ffn

	or

	> orf-purger.pl -f seq1.ffn,seq2.ffn,seq3.ffn,...

   If the sequence file(s) is (are) not in the working directory, the input-gene-files can be
   input as

	> orf-purger.pl -f dir-path/seq.ffn

	or

	> orf-purger.pl -f dir-path/seq1.ffn,dir-path/seq2.ffn,dir-path/seq3.ffn,...

   Note: The filename extension can be any FASTA format: .ffn, .fas, .fsa, .fa, .fasta, or none.

   With Mode 2, the filename-list file can be constructed using the unix command 'ls > list-file-name'
   and edited. Filename entries in a list file contain no path components but the list file must be in
   the same directory as the input sequence files, which enables that directory path to be reflected
   from the pathname of a list file specification.

   Examples for Mode 2:

   If the list file and sequence files are in the working directory from which the program is invoked,
   the input-gene-files can be input as

	> orf-purger.pl -f seq-filenames.lst

   If the list file and sequence files are not in the working directory, the input-gene-files can be
   input as

	> orf-purger.pl -f dir-path/seq-filenames.lst

   Note: The filename extension can be any: .lst, .txt, .tdt, or nothing.

   The beauty to use Mode 2 is manipulating the input order of sequence files as you want from within
   the list file and creating/editing that list file just once but for all repeated running of program
   with such a customized order of input sequence files and without repeated filename-group typing work.

   With Mode 3, all you need to input is just the pathname of a directory where all sequence files are
   stored only (No other irrelevant files allowed).

   An example for Mode 3:

	> orf-purger.pl -f dir-path

   Note: All files are input in the order of sorted names in the directory, which is uncustomizable.

-o <output_genes_file>

   refers to a FASTA file containing all output gene sequences, which can be user-specified.

   * These output gene sequences may or may not contain start and/or stop codons.
   * This output file is used as the submission file for generating oligo probes.

-O <output-overlaps-file>

   refers to a sequence file of Tab-delimited text type (*.tdt), which stores those ORFs overlapping
   with other ORFs (being a subsequence of another gene sequence). These ORF sequences have no oligo
   constructible since every oligo-long segment along such a sequence falls within the longer gene
   sequence with which it overlaps. However an overlap sequence should be used to check if an oligo
   searched from that longer gene sequence won't fall within it. That is, that oligo is constructed
   such that it won't overlap even this discarded ORF, thereby minimizing a potential risk of cross-
   hybridization provided that an ORF dropped as an overlap could exist as a real gene. Therefore,
   these overlap sequences are saved in this file and can be used as a data input for this purpose.
   In this output file, the first column of data is the orf/gene id for the longer sequence and the
   second column is the sequence for the shorter orf/gene (ie, an overlap sequence or an overlap).

-s <orf_stats_file>

   is a file containing a summary of statistics after sequence purging, which is specifiable.

-t <orf_tags_file>

   is a file tagging some particular results during sequence purging, which is specifiable.

-p <orf_ids_prefix>

   refers to a user-provided word code that is used to construct an orf/gene/oligo ID for input gene
   identification. The default prefix uses organism-name identities extracted from the header lines
   of a FASTA sequence file. The length of a prefix must not be more than 10 characters, as the total
   length of a constructed orf/gene/oligo ID has been restricted up to 15 characters.

-b <orf_start_codons>

   enumerates selectable orf starting codons. It is a listing of start codons used to define a working
   orf and has the same formatting rules as the file name listing (ie, comma-separated).

-c <orf_components>

   refers to orf segmentation for the purposes of constructing quality oligo probes.

   The orf-components values are defined as

   1 (component)  -- an orf with both start and stop codons being excluded,
   2 (components) -- an orf with start codon included only, and
   3 (components) -- an orf with both start and stop codons being included.

   The default is 1 (component) that has the highest quality of oligo probes constructed.

-m <minimum_orf_length>

   refers to a minimal length requirement for an orf to be qualified as an entry.

   The default is 66 relative to oligo length of 60, implying that a participating orf should have
   at least the length of summation of oligo, start-codon, and stop-codon lengths. This is the most
   conservative criterion but is specifiable for the least stringent threshold. Some of the suggested
   minimum lengths would cut off at 100-110.

-l <oligo_probe_length>

   is the length of oligo to be generated. The default is 60 bases (60mer).

-r <remove_similarity>

   refers to a sequence purgation based on the similarity between two orfs.

   When the sequences of a genome are pooled over different sources of data such as from strain mutants,
   the chances are those orf sequences are referring to the same gene if their lengths are equal. The
   greater their mutual similarity (ie, the higher percentage of identical bases), the more likely they
   are derived from the originally common ancestor. They diverged from each other just because of one or
   more-than-one-time transition and transversion mutations between individual bases in the evolutionary
   process or simply due to errors introduced throughout the processes of genome sequencing, sequence/
   contig assembly, and gene prediction. Namely, these orfs are considered multiple copies of one gene
   without substantial differentiations between them.

   In terms of construction of oligos from these sequences, however, there is an issue on redundancy of
   such 'mutants', which results in development of much less clusters of homologous sequences and hence
   a dramatic reduction in the number of oligos (each cluster is considered to represent a single gene,
   from which one or more oligo(s) is/are searched for and produced). As such, removal of similar orfs
   becomes necessary and is performed by sequence alignment against a cutoff point in order to yield a
   maximal number of oligos. The default cutoff value is 0.01; that is, just one mismatched base out of
   100 bases is not treated as a difference between two orf sequences, thereby taking one off the gene
   pool and just keeping one singleton is warranted and rendered. This cutoff value should be scaled up
   with a gain in the number of oligos that can be generated by an oligo creating software, say, eArray,
   until that number stops incrementing. This incremented number of oligos may be gainable by combining
   the sequence homology elimination introduced as follows.

-e <eliminate_homology>

   refers to a sequence purgation based on the homology between two orfs.

   In case the above similarity-based purgation does not improve a lot the productivity of oligos, the
   next purgation will be based on the intergenic homology/consensus that would be responsible for less
   clusters of homologous sequences. As described in similarity case, different orfs with equal lengths
   are most likely to be referring to the same gene since the probability of equal-length genes is very
   small. However, it is also possible for unequal-length genes to refer to the same gene due to one or
   more-than-one-time base addition and deletion mutations in the lengthy evolutionary process and again
   a discrepancy caused by errors introduced throughout the processes of genome sequencing, sequence/
   contig assembly, and gene prediction. That is, these orfs are also considered multiple copies of one
   gene without substantial differentiations between them. On the other hand, whatever their lengths might
   be, the orf sequences of this type can be regarded as homologs that also contribute to a small number
   of oligos as long as they have shown a strong homology among them, and hence they are eliminable in
   order to maximize the number of oligos.

   In the light of justification above, sequences are analyzed by local alignment for homolog detection.
   A sequence alignment begins from the first bases following start codons of pairwise orfs and ends at
   the last base of whichever of paired orfs has a shorter length. If a homolog is found, the shorter orf
   is dropped out of gene pool. The default cutoff value to drop a homolog is 0.0 for the initially least
   stringent criterion -- ie, nothing eliminated.

   An example for a scenario where more switches/options are used:

   > orf-purger.pl -i2 -fseq-filenames.lst -oorganism_name.ffn -r.02 -e.01


Note: Running the program in the background is available for freeing up your unix terminal for other tasks.
-----------------------------------------------------------------------------------------------------------
After the program is launched, it will present messages with a prompt asking for one of three options y/b/n.
Typing 'b', and then 'bg', enables you to background the job. When the job is done, type 'fg' to bring back
the foreground and exit.\n\n";
my ($b,$e,$i,$j,$k,$l,$m,$n,$p,$s,$ok,$dir,@files,@tag,@gid,@acn,@gsb,@loc,@prd,@seq);

# Defaults
my $inpFile="input-genes.ffn";
my $outFile="output-purged-genes.ffn";
my $olpFile="output-overlap-genes.tdt";
my $orfStat="orf-stats.txt";
my $orfTags="orf-tags.txt";
my $startcd="atg,gtg,ttg,ctg,att";
my $gprefix="none";
my $inpMode=2;
my $orfComp=1;
my $morfLen=66;
my $oligLen=60;
my $simCull=0.01;
my $homElim=0.0;

unless (@ARGV) {}
else {while ($_=shift) {
if (/^-[cerfiblmphostO]/) {
if ($_ eq "-f") {$inpFile=(defined(local $_=shift)?$_:"");} elsif (/^-f/) {($inpFile)=/-f(.+)/;}
if ($_ eq "-o") {$outFile=(defined(local $_=shift)?$_:"");} elsif (/^-o/) {($outFile)=/-o(.+)/;}
if ($_ eq "-O") {$olpFile=(defined(local $_=shift)?$_:"");} elsif (/^-O/) {($olpFile)=/-O(.+)/;}
if ($_ eq "-s") {$orfStat=(defined(local $_=shift)?$_:"");} elsif (/^-s/) {($orfStat)=/-s(.+)/;}
if ($_ eq "-t") {$orfTags=(defined(local $_=shift)?$_:"");} elsif (/^-t/) {($orfTags)=/-t(.+)/;}
if ($_ eq "-b") {$startcd=(defined(local $_=shift)?$_:"");} elsif (/^-b/) {($startcd)=/-b(.+)/;}
if ($_ eq "-p") {$gprefix=(defined(local $_=shift)?$_:"");} elsif (/^-p/) {($gprefix)=/-p(.+)/;}
if ($_ eq "-c") {$orfComp=(defined(local $_=shift)?$_:0);} elsif (/^-c/) {($orfComp)=/-c(.+)/;}
if ($_ eq "-m") {$morfLen=(defined(local $_=shift)?$_:0);} elsif (/^-m/) {($morfLen)=/-m(.+)/;}
if ($_ eq "-l") {$oligLen=(defined(local $_=shift)?$_:0);} elsif (/^-l/) {($oligLen)=/-l(.+)/;}
if ($_ eq "-r") {$simCull=(defined(local $_=shift)?$_:0);} elsif (/^-r/) {($simCull)=/-r(.+)/;}
if ($_ eq "-e") {$homElim=(defined(local $_=shift)?$_:0);} elsif (/^-e/) {($homElim)=/-e(.+)/;}
if ($_ eq "-i") {$inpMode=(defined(local $_=shift)?$_:0);} elsif (/^-i/) {($inpMode)=/-i(.+)/;}
if ($_ eq "-h") {print "$usage"; exit;}
} else {die "\nUnknown argument/switch $_\n$usage";}}}
if ($inpFile!~/^[\w\s\/\(\)\.,-]*[\w\)]$/) {die "$inpFile: Invalid file name!\n";}
if ($startcd!~/^[acgt,]+[acgt]$/i) {die "$startcd: Invalid start codons!\n";}
if ($orfComp!~/^\d+$/) {die "$orfComp: Not numerical parameter!\n";}
if ($morfLen!~/^\d+$/) {die "$morfLen: Not numerical parameter!\n";}
if ($oligLen!~/^\d+$/) {die "$oligLen: Not numerical parameter!\n";}
if ($simCull!~/^0?\.?\d*$/) {die "$simCull: Not numerical/decimal parameter!\n";}
if ($homElim!~/^0?\.?\d*$/) {die "$homElim: Not numerical/decimal parameter!\n";}
if ($inpMode!~/^[123]$/) {die "$inpMode: Not a 1-3 number!\n";}

print "\nThis program will run based on the following settings:\n
-i input_file_mode    = $inpMode
-f input_gene_files   = $inpFile
-o output_genes_file  = $outFile
-O out_overlaps_file  = $olpFile
-s orf_stats_file     = $orfStat
-t orf_tags_file      = $orfTags
-p orf_ids_prefix     = $gprefix
-b orf_start_codons   = $startcd
-c orf_components     = $orfComp
-m minimum_orf_length = $morfLen
-l oligo_probe_length = $oligLen
-r remove_similarity  = $simCull
-e eliminate_homology = $homElim\n
Type in <$0 -h> for help and make your own settings.\n\n";

do {
	print "Proceed to run ? [y/b/n]: ";
	$b = <STDIN>;
	if ($b=~/^n$/i) {exit;}
} until ($b=~/^[yb]$/i);
if ($b=~/b/i) {
	print "\nPlease type bg to run in the background.\n";
	kill 'TSTP', $$;
}
if ($gprefix eq "none") {$gprefix="";}
$startcd=uc($startcd);
$startcd=~s/,\s*/|^/g;
print "\n";

if ($inpMode==1) {
if (-d $inpFile) {
	die "$inpFile: This is a directory, not file(s).\n";
}} elsif ($inpMode==2) {
	unless (-f $inpFile) {die "$inpFile: Not a file.\n";}
	$dir=$inpFile=~/\//?dirname($inpFile).'/':""; # save dir path
	open(INPFILE, "<$inpFile") or die "$inpFile: No such a file.\n";
	@files=<INPFILE>;
	$inpFile=join('',@files);
	$inpFile=~s/\n/,/g;
	close(INPFILE);
} elsif ($inpMode==3) {
	opendir(dH, $inpFile) || die "can't open directory $inpFile: $!";
	# get all file entries in the directory w/o dots (.) and (..)
	@files=grep(!/^\./, readdir(dH));
	@files=sort(@files);
	$dir="$inpFile/"; # save dir path
	$inpFile=join(',',@files);
	closedir(dH);
}

open (OUTFILE, ">$outFile") or die "$!";
open (OVRLAPS, ">$olpFile") or die "$!";
open (OUTSTAT, ">$orfStat") or die "$!";
open (OUTTAGS, ">$orfTags") or die "$!";

local $/=">"; # set to fasta input record separator
print OUTTAGS "Information of identical sequences found and purged (2nd one of pairwise orfs is removed):\n";
print OUTTAGS "==========================================================================================\n\n";
$e=$i=$m=$n=$p=$ok=0;
while ($inpFile=~/([\w\s\/\(\)\.-]*[\w\)])/g) {
	print "Processing $1...\n";
	open (INPFILE, "<".($inpMode==1?$1:$dir.$1)) or die "$1: No such a file.\n";
	if ($1=~/[\.\s_-]([^\.\s_-]+)\.\w{2,5}$/) {} # grab last part of a filename
	elsif ($1=~/[\.\s_-]([^\.\s_-]+)$/) {}
	elsif ($1=~/(.{4})$/) {}
	@files=$1;
while (<INPFILE>) {
	next if (/^\s*>\s*$/); # skip the very beginning ">"
	chomp; # trim the trailing input record separator ">"
	($s)=/^\s*(.+)/;  # extract fasta header line
	if ($s=~/^\s*\w+\|/) { # from GenBank databases
		if ($s=~/^\s*\w+\|(.+)\|:(\w??)(\d+)-(\d+)\s+(.+)\s+(\w{3,}_\w*\d{4})\s+\[.+\s(\S+)\]/) {}
		elsif ($s=~/^\s*\w+\|(.+)\|:(\w??)(\d+)-(\d+)\s+(.+)\s+()\[.+\s(\S+)\]/) {} # no gene symbol
		elsif ($s=~/^\s*\w+\|(.+)\|:(\w??)(\d+)-(\d+)\s+()()\[.+\s(\S+)\]/) {} # no gene symbol & product name
		elsif ($s=~/^\s*\w+\|(.+)\|:(\w??)(\d+)-(\d+)()()()/) {} # no gene symbol, product name & organism name
		$gid[$e]=($gprefix eq ""?$7:$gprefix).'_'.sprintf("%04d",$e+1); # construct an orf/gene/oligo ID
		$acn[$e]=$1;
		$gsb[$e]=$6 eq ""?"gene_symbol":$6;
		$loc[$e]=$2 eq 'c'?"$3..$4(-)":"$3..$4(+)";
		$prd[$e]=$5 eq ""?"product_name":$5;
	} else { # from JGI and other databases
		if ($s=~/^\s*(\S+)\s+(\w{3,}_\w*\d{4})\s+(.+)\s+(\d+\.\.\d+\([+-]\))/) {}
		elsif ($s=~/^\s*(\S+)\s+(\w{3,}_\w*\d{4})\s+()(\d+\.\.\d+\([+-]\))/) {} # no product name
		elsif ($s=~/^\s*(\S+)\s+(\w{3,}_\w*\d{4})()()/) {} # no product name and gene locus
		elsif ($s=~/^\s*(\S+)()()()/) {} # no gene symbol, product name and gene locus
		$gid[$e]=$gprefix eq ""?$1:$gprefix.'_'.sprintf("%04d",$e+1); # construct an orf/gene/oligo ID
		$acn[$e]="gene_accession";
		$gsb[$e]=$2 eq ""?"gene_symbol":$2;
		$loc[$e]=$4 eq ""?"gene_locus":$4;
		$prd[$e]=$3 eq ""?"product_name":$3;
	}
	s/.+\n//; # strip off the 1st line of sequence block
	tr/acgt/ACGT/; # uppercasing bases of sequence block
	s/\s//g;  # strip whitespace (newline etc) off block
	if ($e > 0) {
		# Detect duplicate sequences
		for ($i=$e-1; $i>=0; $i--) {
		last if ($_ eq $seq[$i]);
	}}
	# Remove duplicate sequences by storing unique orfs only
	if ($i<0 or $e==0) { # has distinct orfs
		if ($gsb[$e]=~/gene_symbol/&&$loc[$e]=~/gene_locus/&&$prd[$e]=~/product_name/) {
		$gsb[$e]=$gid[$e];
		$gid[$e]=$files[0].'_'.sprintf("%04d",$e+1); # customize an orf/gene/oligo ID
		}
		$seq[$e]=$_; # save sequence string
		$ok += 1;  # count distinct orfs
		$e += 1; # increment array size
	} else { # duplicate orfs hit
		# integrate fasta header line information
		if ($loc[$i]=~/gene_locus/&&$loc[$e]!~/gene_locus/) {$loc[$i]=$loc[$e];}
		if ($acn[$i]=~/gene_accession/&&$acn[$e]!~/gene_accession/) {$acn[$i]=$acn[$e];}
		if ($gsb[$i]=~/gene_symbol/&&$gsb[$e]!~/gene_symbol/||($gsb[$i]!~/^\w{3,}_\w*\d{4}$/&&$gsb[$e]=~/^\w{3,}_\w*\d{4}$/)) {$gsb[$i]=$gsb[$e];}
		if ($prd[$i]=~/product_name/&&$prd[$e]!~/product_name/||($prd[$i]=~/hypothetical/i&&$prd[$e]!~/hypothetical/i)) {$prd[$i]=$prd[$e];}
		elsif ($prd[$e]!~/hypothetical/i&&$prd[$e]!~/product_name/) {
		($s=lc($prd[$i]))=~s/[-,\s]//g; # strip differential characters off product names
		($_=lc($prd[$e]))=~s/[-,\s]//g; # and lowercase them to make them comparable
		if ($_ ne $s) {
			if (index($_,$s)>=0) {$prd[$i]=$prd[$e];} # if $_ contains $s
			elsif (index($s,$_)>=0) {} # do nothing if $s contains $_
			else {$prd[$i].='|'.$prd[$e];} # merge info delimited by |
		}
	}
	print OUTTAGS "$gid[$i]\t$acn[$i]\t$gsb[$i]\t$loc[$i]\t$prd[$i]\n";
	print OUTTAGS "$gid[$e]\t$acn[$e]\t$gsb[$e]\t$loc[$e]\t$prd[$e]\n\n";
	$n += 1; # count duplicates
	}}
	$p += 1; # count input files
	$m += $.-1; # count input records
	close(INPFILE); # reset $.
}

print OUTSTAT "\nNumber of input files = $p\n";
print OUTSTAT "Number of input sequences = $m\n\n";
print OUTSTAT "Number of duplicate sequences = $n\n";
print OUTSTAT "Number of distinct sequences = $ok\n\n";
print "\nNumber of input files = $p\n";
print "Number of input sequences = $m\n";
print "Number of duplicate sequences = $n\n";
print "Number of distinct sequences = $ok\n\n";
print OUTTAGS "\n\n";

$/="\n"; # reset to normal input record separator
# Remove pseudo-genes and get real genes/orfs
print "Checking component sequences...\n";
$n=$k=$ok=0;
for ($i=$#seq; $i>=0; $i--) {
	if ($seq[$i]=~/^$startcd/g) {
	# a start codon matched
	$m=0; # initialize status
	while ($seq[$i]=~/(.{3})/g) {
	$s=$1;
	if ($1=~/TA[AG]|TGA/g) {
	$m=1; # a stop codon matched
	if (pos($seq[$i]) < $morfLen) {
	print OUTTAGS "ORF $gid[$i]: Shorter than oligo length ($oligLen bp)! Sequence length: ".length($seq[$i])."  Stop codon: $s  In-frame length: ".pos($seq[$i])." < $morfLen\n";
	print OUTTAGS "$seq[$i]\n\n";
	splice(@gid,$i,1);
	splice(@acn,$i,1);
	splice(@gsb,$i,1);
	splice(@loc,$i,1);
	splice(@prd,$i,1);
	splice(@seq,$i,1);
	$k += 1;
	} else { # has required in-frame length
	$ok += 1;
	if ($orfComp==1) {
		$seq[$i]=substr($seq[$i],3,pos($seq[$i])-6); # get orf w/o start and stop codon
	} elsif ($orfComp==2) {
		$seq[$i]=substr($seq[$i],0,pos($seq[$i])-3); # get orf w/ start codon only
	} elsif ($orfComp==3) {
		$seq[$i]=substr($seq[$i],0,pos($seq[$i])); # get orf w/ both start and stop codon
	}}
	last;
	}}
	unless ($m) { # if stop codon not matched
		print OUTTAGS "ORF $gid[$i]: Stop codon not found!\n";
		print OUTTAGS "$seq[$i]\n\n";
		splice(@gid,$i,1);
		splice(@acn,$i,1);
		splice(@gsb,$i,1);
		splice(@loc,$i,1);
		splice(@prd,$i,1);
		splice(@seq,$i,1);
		$n += 1;
	}} else { # if start codon not matched
		print OUTTAGS "ORF $gid[$i]: Start codon not found!\n";
		print OUTTAGS "$seq[$i]\n\n";
		splice(@gid,$i,1);
		splice(@acn,$i,1);
		splice(@gsb,$i,1);
		splice(@loc,$i,1);
		splice(@prd,$i,1);
		splice(@seq,$i,1);
		$n += 1;
	}
}

print OUTSTAT "Number of genes < $oligLen bp = $k\n";
print OUTSTAT "Number of pseudo-genes = $n\n";
print OUTSTAT "Number of potential genes = $ok\n\n";
print OUTTAGS "\n\n";

# Remove identical orfs/genes
print "Checking identical component sequences...\n\n";
print OUTTAGS "Information of identical component sequences found and purged (2nd one of pairwise sequences is removed):\n";
print OUTTAGS "=========================================================================================================\n\n";
$n=$ok=0;
for ($i=$#seq; $i>=0; $i--) {
	if ($i==0) {$ok += 1; last;}
	for ($j=$i-1; $j>=0; $j--) {
	if ($seq[$i] eq $seq[$j]) {
		# duplicate orfs hit
		# integrate fasta header line information
		if ($loc[$j]=~/gene_locus/&&$loc[$i]!~/gene_locus/) {$loc[$j]=$loc[$i];}
		if ($acn[$j]=~/gene_accession/&&$acn[$i]!~/gene_accession/) {$acn[$j]=$acn[$i];}
		if ($gsb[$j]=~/gene_symbol/&&$gsb[$i]!~/gene_symbol/||($gsb[$j]!~/^\w{3,}_\w*\d{4}$/&&$gsb[$i]=~/^\w{3,}_\w*\d{4}$/)) {$gsb[$j]=$gsb[$i];}
		if ($prd[$j]=~/product_name/&&$prd[$i]!~/product_name/||($prd[$j]=~/hypothetical/i&&$prd[$i]!~/hypothetical/i)) {$prd[$j]=$prd[$i];}
		elsif ($prd[$i]!~/hypothetical/i&&$prd[$i]!~/product_name/) {
		($s=lc($prd[$j]))=~s/[-,\s]//g; # strip differential/noisy characters off product names
		($_=lc($prd[$i]))=~s/[-,\s]//g; # and lowercase them to make them comparable
		if ($_ ne $s) {
			if (index($_,$s)>=0) {$prd[$j]=$prd[$i];} # if $_ contains $s
			elsif (index($s,$_)>=0) {} # do nothing if $s contains $_
			else {$prd[$j].='|'.$prd[$i];} # merge info delimited by |
		}}
		print OUTTAGS "$gid[$j]\t$acn[$j]\t$gsb[$j]\t$loc[$j]\t$prd[$j]\n";
		print OUTTAGS "$gid[$i]\t$acn[$i]\t$gsb[$i]\t$loc[$i]\t$prd[$i]\n\n";
		splice(@gid,$i,1);
		splice(@acn,$i,1);
		splice(@gsb,$i,1);
		splice(@loc,$i,1);
		splice(@prd,$i,1);
		splice(@seq,$i,1);
		$n += 1;
		last;
	}}
	if ($j<0) {$ok += 1;}
}

print OUTSTAT "Number of identical genes = $n\n";
print OUTSTAT "Number of distinct genes = $ok\n\n";

# Remove overlapping genes
print OUTTAGS "\nIdentical subsequences found and purged (2nd one of pairwise ORFs is an overlap and gets removed):\n";
print OUTTAGS "==================================================================================================\n\n";
$n=$ok=0;
for ($i=$#seq; $i>=0; $i--) {
	print "Checking overlap ORF ",$i+1,"...\n";
	if ($i==0) {$ok += 1; last;}
	$l=length($seq[$i]);
	for ($j=$i-1; $j>=0; $j--) {
		if ($l <= length($seq[$j])) {
			if ($seq[$j]=~/$seq[$i]/) {
				print OUTTAGS "$gid[$j]\t$seq[$j]\n";
				print OUTTAGS "$gid[$i]\t$seq[$i]\n\n";
				print OVRLAPS "$gid[$j]\t$seq[$i]\n";
				splice(@gid,$i,1);
				splice(@acn,$i,1);
				splice(@gsb,$i,1);
				splice(@loc,$i,1);
				splice(@prd,$i,1);
				splice(@seq,$i,1);
				$n += 1;
				last;
		}} else {
			if ($seq[$i]=~/$seq[$j]/) {
				print OUTTAGS "$gid[$i]\t$seq[$i]\n";
				print OUTTAGS "$gid[$j]\t$seq[$j]\n\n";
				print OVRLAPS "$gid[$i]\t$seq[$j]\n";
				splice(@gid,$j,1);
				splice(@acn,$j,1);
				splice(@gsb,$j,1);
				splice(@loc,$j,1);
				splice(@prd,$j,1);
				splice(@seq,$j,1);
				$i -= 1;
				$n += 1;
		}}
	}
	if ($j<0) {$ok += 1;}
}

print OUTSTAT "Number of overlapping genes = $n\n";
print OUTSTAT "Number of non-overlap genes = $ok\n\n";

# Remove similar orfs by local alignment
if ($simCull > 0) {
print OUTTAGS "\nSimilar sequences found and purged (2nd one of pairwise ORFs has a similarity and is removed; l=length of either ORF, n=number of unmatched bases, pct=percentage of mismatches):\n";
print OUTTAGS "=================================================================================================================================================================================\n\n";
$n=$ok=0;
for ($i=$#seq; $i>=0; $i--) {
	print "Checking similar ORF ",$i+1,"...\n";
	if ($i==0) {$ok += 1; last;}
	$l=length($seq[$i]);
	for ($j=$i-1; $j>=0; $j--) {
		if ($l==length($seq[$j])) {
			$m=0;
			for ($k=0; $k<$l; $k++) {
				if (substr($seq[$i],$k,1) ne substr($seq[$j],$k,1)) {
					$m += 1;
				}
			}
			$p = $m/$l;
			if ($p <= $simCull) {
				print OUTTAGS "l=$l\tn=$m\tpct=$p\n";
				print OUTTAGS "$gid[$j]\t$seq[$j]\n";
				print OUTTAGS "$gid[$i]\t$seq[$i]\n\n";
				splice(@gid,$i,1);
				splice(@acn,$i,1);
				splice(@gsb,$i,1);
				splice(@loc,$i,1);
				splice(@prd,$i,1);
				splice(@seq,$i,1);
				$n += 1;
				last;
			}
		}
	}
	if ($j<0) {$ok += 1;}
	}
	print OUTSTAT "Number of similar genes = $n\n";
	print OUTSTAT "Number of singleton genes = $ok\n\n";
}

# Eliminate homologous genes
if ($homElim > 0) {
print OUTTAGS "\nHomologous sequences found and purged (2nd one of pairwise ORFs is eliminated; l=length of shorter ORF, n=number of unmatched bases, pct=percentage of mismatches):\n";
print OUTTAGS "===================================================================================================================================================================\n\n";
$n=$ok=0;
for ($i=$#seq; $i>=0; $i--) {
	print "Checking homolog ORF ",$i+1,"...\n";
	if ($i==0) {$ok += 1; last;}
	$l=length($seq[$i]);
	for ($j=$i-1; $j>=0; $j--) {
		$m=length($seq[$j]);
		if ($l < $m) {
			$e=0;
			for ($k=0; $k<$l; $k++) {
				if (substr($seq[$i],$k,1) ne substr($seq[$j],$k,1)) {
					$e += 1;
				}
			}
			$p = $e/$l;
			if ($p <= $homElim) {
				print OUTTAGS "l=$l\tn=$e\tpct=$p\n";
				print OUTTAGS "$gid[$j]\t$seq[$j]\n";
				print OUTTAGS "$gid[$i]\t$seq[$i]\n\n";
				splice(@gid,$i,1);
				splice(@acn,$i,1);
				splice(@gsb,$i,1);
				splice(@loc,$i,1);
				splice(@prd,$i,1);
				splice(@seq,$i,1);
				$n += 1;
				last;
			}
		} elsif ($l > $m) {
			$e=0;
			for ($k=0; $k<$m; $k++) {
				if (substr($seq[$i],$k,1) ne substr($seq[$j],$k,1)) {
					$e += 1;
				}
			}
			$p = $e/$m;
			if ($p <= $homElim) {
				print OUTTAGS "l=$m\tn=$e\tpct=$p\n";
				print OUTTAGS "$gid[$i]\t$seq[$i]\n";
				print OUTTAGS "$gid[$j]\t$seq[$j]\n\n";
				splice(@gid,$j,1);
				splice(@acn,$j,1);
				splice(@gsb,$j,1);
				splice(@loc,$j,1);
				splice(@prd,$j,1);
				splice(@seq,$j,1);
				$i -= 1;
				$n += 1;
			}
		}
	}
	if ($j<0) {$ok += 1;}
	}
	print OUTSTAT "Number of homologous genes = $n\n";
	print OUTSTAT "Number of non-homolog genes = $ok\n\n";
}

# Check if an oligo exist across all genes
$e=$#seq;
$n=$ok=0;
for ($i=$e; $i>=0; $i--) {
	print "Checking if ORF# ",$i+1," has an oligo...\n";
	$l=length($seq[$i])-$oligLen;
	for ($j=0; $j<=$l; $j++) {
		$s=substr($seq[$i],$j,$oligLen);
		for ($k=$e; $k>=0; $k--) {
			if ($k != $i) {
				last if ($seq[$k]=~/$s/);
			}
		}
		if ($k < 0) {
			print "ORF $gid[$i] has oligo.\n";
			$ok += 1;
			last;
		}
	}
	if ($j > $l) {
		print "ORF $gid[$i] has no oligo.\n";
		splice(@gid,$i,1);
		splice(@acn,$i,1);
		splice(@gsb,$i,1);
		splice(@loc,$i,1);
		splice(@prd,$i,1);
		splice(@seq,$i,1);
		$e -= 1;
		$n += 1;
	}
}

print OUTSTAT "Number of genes w/o oligos available = $n\n";
print OUTSTAT "Number of genes w/ oligos available = $ok\n\n";

# Output quality sequences
$n=0;
$e=$#seq;
for ($i=0; $i<=$e; $i++) {
	print OUTFILE ">$gid[$i] $acn[$i] $gsb[$i] $loc[$i] $prd[$i]\n";
	$m=length($seq[$i]);
	# Print sequence in lines of 60 bp
	for ($j=0; $j < $m; $j += 60) {
		print OUTFILE substr($seq[$i],$j,60),"\n";
	}
	$n += 1;
}

print OUTSTAT "Number of output unique genes = $n\n";
print "\nNumber of output unique genes = $n\n\n";

close(OUTFILE);
close(OUTSTAT);
close(OUTTAGS);
print "Done.\n\n";
if ($b=~/b/i) {
	print "Please type fg to return to the foreground and exit.\n";
}
